@@ -29,7 +29,7 @@
 #include <osmocom/msc/e_link.h>
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/msc_roles.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/ran_infra.h>
 #include <osmocom/msc/msc_a.h>
 #include <osmocom/msc/msc_a_remote.h>
