@@ -41,6 +41,7 @@
 #include <osmocom/msc/msc_t.h>
 #include <osmocom/msc/call_leg.h>
 #include <osmocom/msc/rtp_stream.h>
+#include <osmocom/msc/codec_mapping.h>
 
 #include "msc_vlr_tests.h"
 
@@ -76,6 +77,7 @@ bool bssap_clear_sent = false;
 
 bool bssap_assignment_expected = false;
 bool bssap_assignment_sent = false;
+struct gsm0808_channel_type bssap_assignment_command_last_channel_type;
 bool iu_rab_assignment_expected = false;
 bool iu_rab_assignment_sent = false;
 
@@ -83,9 +85,10 @@ uint32_t cc_to_mncc_tx_expected_msg_type = 0;
 const char *cc_to_mncc_tx_expected_imsi = NULL;
 bool cc_to_mncc_tx_confirmed = false;
 uint32_t cc_to_mncc_tx_got_callref = 0;
+char cc_to_mncc_tx_last_sdp[1024] = {};
 
-enum rtp_direction expecting_crcx = -1;
-bool got_crcx = false;
+bool expecting_crcx[2] = {};
+bool got_crcx[2] = {};
 
 extern int ran_dec_dtap_undup_pdisc_ctr_bin(uint8_t pdisc);
 
@@ -307,14 +310,18 @@ static int bssap_validate_cipher_mode_cmd(const struct ran_cipher_mode_command *
 	return 0;
 }
 
-static void bssap_validate_assignment_cmd()
+static void bssap_validate_assignment_cmd(const struct ran_assignment_command *assignment_command)
 {
 	OSMO_ASSERT(bssap_assignment_expected);
 	bssap_assignment_expected = false;
 	bssap_assignment_sent = true;
+	if (assignment_command->channel_type)
+		bssap_assignment_command_last_channel_type = *assignment_command->channel_type;
+	else
+		bssap_assignment_command_last_channel_type = (struct gsm0808_channel_type){};
 }
 
-static void iucs_validate_assignment_cmd()
+static void iucs_validate_assignment_cmd(const struct ran_assignment_command *assignment_command)
 {
 	OSMO_ASSERT(iu_rab_assignment_expected);
 	iu_rab_assignment_expected = false;
@@ -374,10 +381,10 @@ struct msgb *dont_ran_encode(struct osmo_fsm_inst *caller_fi, const struct ran_m
 	case RAN_MSG_ASSIGNMENT_COMMAND:
 		switch (ran_type) {
 		case OSMO_RAT_GERAN_A:
-			bssap_validate_assignment_cmd();
+			bssap_validate_assignment_cmd(&ran_enc_msg->assignment_command);
 			break;
 		case OSMO_RAT_UTRAN_IU:
-			iucs_validate_assignment_cmd();
+			iucs_validate_assignment_cmd(&ran_enc_msg->assignment_command);
 			break;
 		default:
 			OSMO_ASSERT(false);
@@ -408,6 +415,16 @@ struct ran_infra test_ran_infra[] = {
 		.log_subsys = DIUCS,
 		.tdefs = msc_tdefs_utran,
 		.ran_encode = dont_ran_encode,
+		.force_mgw_codecs_to_ran = {
+			.count = 1,
+			.codec = {
+				{
+					.payload_type = 96,
+					.subtype_name = "VND.3GPP.IUFP",
+					.rate = 16000,
+				},
+			},
+		},
 	},
 };
 
@@ -419,7 +436,7 @@ static int fake_msc_a_ran_dec(const struct ran_msg *ran_dec_msg)
 	return msc_a_ran_decode_cb(g_msub->role[MSC_ROLE_A], &d, ran_dec_msg);
 }
 
-void rx_from_ms(struct msgb *msg)
+void rx_from_ms(struct msgb *msg, const struct gsm0808_speech_codec_list *codec_list_bss_supported)
 {
 	struct gsm48_hdr *gh = msgb_l3(msg);
 	struct ran_msg ran_dec_msg;
@@ -452,6 +469,7 @@ void rx_from_ms(struct msgb *msg)
 			.compl_l3 = {
 				.cell_id = &cell_id,
 				.msg = msg,
+				.codec_list_bss_supported = codec_list_bss_supported,
 			},
 		};
 	} else {
@@ -479,7 +497,30 @@ void ms_sends_msg(const char *hex)
 
 	msg = msgb_from_hex("ms_sends_msg", 1024, hex);
 	msg->l1h = msg->l2h = msg->l3h = msg->data;
-	rx_from_ms(msg);
+	rx_from_ms(msg, NULL);
+	msgb_free(msg);
+}
+
+void ms_sends_msgf(const char *fmt, ...)
+{
+	va_list ap;
+	char *hex;
+
+	va_start(ap, fmt);
+	hex = talloc_vasprintf(msc_vlr_tests_ctx, fmt, ap);
+	va_end(ap);
+
+	ms_sends_msg(hex);
+	talloc_free(hex);
+}
+
+void ms_sends_compl_l3(const char *hex, const struct gsm0808_speech_codec_list *codec_list_bss_supported)
+{
+	struct msgb *msg;
+
+	msg = msgb_from_hex("ms_sends_msg", 1024, hex);
+	msg->l1h = msg->l2h = msg->l3h = msg->data;
+	rx_from_ms(msg, codec_list_bss_supported);
 	msgb_free(msg);
 }
 
@@ -640,8 +681,8 @@ void clear_vlr()
 
 	osmo_gettimeofday_override = false;
 
-	expecting_crcx = -1;
-	got_crcx = false;
+	memset(expecting_crcx, 0, sizeof(expecting_crcx));
+	memset(got_crcx, 0, sizeof(got_crcx));
 
 	bssap_assignment_expected = false;
 	bssap_assignment_sent = false;
@@ -722,8 +763,18 @@ struct gsm_mncc *on_call_release_mncc_sends_to_cc_data = NULL;
 int mncc_recv(struct gsm_network *net, struct msgb *msg)
 {
 	struct gsm_mncc *mncc = (void*)msg->data;
-	log("MSC --> MNCC: callref 0x%x: %s", mncc->callref,
-	    get_mncc_name(mncc->msg_type));
+	if (mncc->msg_type == MNCC_RTP_CREATE) {
+		struct gsm_mncc_rtp *rtp = (void *)msg->data;
+		log("MSC --> MNCC: callref 0x%x: %s\n%s", rtp->callref,
+		    get_mncc_name(rtp->msg_type),
+		    rtp->sdp);
+		OSMO_STRLCPY_ARRAY(cc_to_mncc_tx_last_sdp, rtp->sdp);
+	} else {
+		log("MSC --> MNCC: callref 0x%x: %s\n%s", mncc->callref,
+		    get_mncc_name(mncc->msg_type),
+		    mncc->sdp);
+		OSMO_STRLCPY_ARRAY(cc_to_mncc_tx_last_sdp, mncc->sdp);
+	}
 
 	if (mncc->msg_type == MNCC_REL_IND && on_call_release_mncc_sends_to_cc_data) {
 
@@ -835,27 +886,35 @@ struct rtp_stream fake_rtp[2] = {
 
 void expect_crcx(enum rtp_direction towards)
 {
-	OSMO_ASSERT(expecting_crcx == -1);
-	expecting_crcx = towards;
-	got_crcx = false;
+	OSMO_ASSERT(!expecting_crcx[towards]);
+	expecting_crcx[towards] = true;
+	got_crcx[towards] = false;
+}
+
+bool crcx_scheduled(enum rtp_direction towards)
+{
+	return got_crcx[towards];
 }
 
 /* override, requires '-Wl,--wrap=call_leg_ensure_ci' */
 int __real_call_leg_ensure_ci(struct call_leg *cl, enum rtp_direction dir, uint32_t call_id, struct gsm_trans *for_trans,
-			      const enum mgcp_codecs *codec_if_known, const struct osmo_sockaddr_str *remote_addr_if_known);
+			      const struct sdp_audio_codecs *codecs_if_known,
+			      const struct osmo_sockaddr_str *remote_addr_if_known);
 int __wrap_call_leg_ensure_ci(struct call_leg *cl, enum rtp_direction dir, uint32_t call_id, struct gsm_trans *for_trans,
-			      const enum mgcp_codecs *codec_if_known, const struct osmo_sockaddr_str *remote_addr_if_known)
+			      const struct sdp_audio_codecs *codecs_if_known,
+			      const struct osmo_sockaddr_str *remote_addr_if_known)
 {
 	if (!cl->rtp[dir]) {
-		log("MGW <--CRCX to %s-- MSC: callref=0x%x", rtp_direction_name(dir), call_id);
+		log("MGW <--CRCX to %s-- MSC: call_id=0x%x codecs=%s", rtp_direction_name(dir), call_id,
+		    codecs_if_known ? sdp_audio_codecs_to_str(codecs_if_known) : "unset");
 
-		OSMO_ASSERT(expecting_crcx == dir);
-		expecting_crcx = -1;
-		got_crcx = true;
+		OSMO_ASSERT(expecting_crcx[dir]);
+		expecting_crcx[dir] = false;
+		got_crcx[dir] = true;
 
 		call_leg_ensure_rtp_alloc(cl, dir, call_id, for_trans);
-		if (codec_if_known)
-			rtp_stream_set_codec(cl->rtp[dir], *codec_if_known);
+		if (codecs_if_known)
+			rtp_stream_set_codecs(cl->rtp[dir], codecs_if_known);
 		if (remote_addr_if_known && osmo_sockaddr_str_is_nonzero(remote_addr_if_known))
 			rtp_stream_set_remote_addr(cl->rtp[dir], remote_addr_if_known);
 	}
@@ -875,7 +934,7 @@ void crcx_ok(enum rtp_direction dir)
 	osmo_fsm_inst_dispatch(cl->fi, CALL_LEG_EV_RTP_STREAM_ADDR_AVAILABLE, cl->rtp[dir]);
 }
 
-static int fake_vlr_tx_lu_acc(void *msc_conn_ref, uint32_t send_tmsi)
+static int fake_vlr_tx_lu_acc(void *msc_conn_ref, uint32_t send_tmsi, enum vlr_lu_type lu_type)
 {
 	struct msc_a *msc_a = msc_conn_ref;
 	if (send_tmsi == GSM_RESERVED_TMSI)
@@ -887,7 +946,7 @@ static int fake_vlr_tx_lu_acc(void *msc_conn_ref, uint32_t send_tmsi)
 	return 0;
 }
 
-static int fake_vlr_tx_lu_rej(void *msc_conn_ref, enum gsm48_reject_value cause)
+static int fake_vlr_tx_lu_rej(void *msc_conn_ref, enum gsm48_reject_value cause, enum vlr_lu_type lu_type)
 {
 	struct msc_a *msc_a = msc_conn_ref;
 	btw("sending LU Reject for %s, cause %u", msc_a->c.fi->id, cause);
@@ -1000,17 +1059,22 @@ void ms_sends_security_mode_complete(uint8_t utran_encryption)
 		g_msub = NULL;
 }
 
-void ms_sends_assignment_complete(enum mgcp_codecs assigned_codec)
+void ms_sends_assignment_complete(const char *sdp_codec_name)
 {
 	struct ran_msg ran_dec;
+	const struct codec_mapping *m = codec_mapping_by_subtype_name(sdp_codec_name);
+	OSMO_ASSERT(m);
+	OSMO_ASSERT(m->has_gsm0808_speech_codec);
 
 	ran_dec = (struct ran_msg){
 		.msg_type = RAN_MSG_ASSIGNMENT_COMPLETE,
 		.assignment_complete = {
 			.codec_present = true,
-			.codec = assigned_codec,
+			.codec = m->gsm0808_speech_codec,
+			.codec_with_iuup = (rx_from_ran == OSMO_RAT_UTRAN_IU),
 		},
 	};
+	osmo_sockaddr_str_from_str(&ran_dec.assignment_complete.remote_rtp, "1.2.3.4", 1234);
 	fake_msc_a_ran_dec(&ran_dec);
 
 	if (!conn_exists(g_msub))
@@ -1171,10 +1235,10 @@ struct gsm_network *test_net(void *ctx)
 	INIT_LLIST_HEAD(&net->iu.sri->ran_conns);
 
 	net->mgw.tdefs = g_mgw_tdefs;
-	mgcp_client_conf_init(&net->mgw.conf);
 	net->mgw.tdefs = g_mgw_tdefs;
+	net->mgw.conf = mgcp_client_conf_alloc(net);
 	net->mgw.mgw_pool = mgcp_client_pool_alloc(net);
-	client = mgcp_client_init(net, &net->mgw.conf);
+	client = mgcp_client_init(net, net->mgw.conf);
 	mgcp_client_pool_register_single(net->mgw.mgw_pool, client);
 	return net;
 }
@@ -1189,6 +1253,9 @@ int main(int argc, char **argv)
 	msgb_ctx = msgb_talloc_ctx_init(msc_vlr_tests_ctx, 0);
 	osmo_init_logging2(msc_vlr_tests_ctx, &info);
 
+	osmo_vlr_set_log_cat(OSMO_VLR_LOGC_VLR, DVLR);
+	osmo_vlr_set_log_cat(OSMO_VLR_LOGC_SGS, DSGS);
+
 	_log_lines = cmdline_opts.verbose;
 
 	OSMO_ASSERT(osmo_stderr_target);
